/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Video4Linux (webcam) image grabber
 *
 * based on vgrabber.c from Phil Blundell <philb@gnu.org>
 *  source:    http://www.tazenda.demon.co.uk/phil/vgrabber.c
 *  some info: http://www.tazenda.demon.co.uk/phil/v4l.html
 * and yuv2ppm.c by Scott Scriven <toykeeper@cheerful.com>
 *  source:    http://www.fourcc.org/yuv2ppm.c
 *  some info: http://www.fourcc.org/fccyvrgb.php
 *
 * modified to support Creative WebCam Go without postprocessing kernel module
 * (see http://www.linux-projects.org/modules/mydownloads/viewcat.php)
 * and allow the setting of image parameters
*/

/* copyright (C) 2005 by Erik Auerswald <auerswal@unix-ag.uni-kl.de> */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <strings.h>
#include <getopt.h>

#include <linux/types.h>
#include <linux/videodev.h>

#define DEFAULT_FILE "/dev/video0"
#define DEFAULT_OUTFILE "-"
#define VERSION "1.1"
#define MAX_BRIGHTNESS_LOOPS 100

/* Stole this from tvset.c */

#define READ_VIDEO_PIXEL(buf, format, depth, r, g, b)			\
{									\
	switch (format)							\
	{								\
		case VIDEO_PALETTE_GREY:				\
			switch (depth)					\
			{						\
				case 4:					\
				case 6:					\
				case 8:					\
					(r) = (g) = (b) = (*buf++ << 8);\
					break;				\
									\
				case 16:				\
					(r) = (g) = (b) = 		\
						*((unsigned short *) buf);	\
					buf += 2;			\
					break;				\
			}						\
			break;						\
									\
									\
		case VIDEO_PALETTE_RGB565:				\
		{							\
			unsigned short tmp = *(unsigned short *)buf;	\
			(r) = tmp&0xF800;				\
			(g) = (tmp<<5)&0xFC00;				\
			(b) = (tmp<<11)&0xF800;				\
			buf += 2;					\
		}							\
		break;							\
									\
		case VIDEO_PALETTE_RGB555:				\
			(r) = (buf[0]&0xF8)<<8;				\
			(g) = ((buf[0] << 5 | buf[1] >> 3)&0xF8)<<8;	\
			(b) = ((buf[1] << 2 ) & 0xF8)<<8;		\
			buf += 2;					\
			break;						\
									\
		case VIDEO_PALETTE_RGB24:				\
			(r) = buf[0] << 8; (g) = buf[1] << 8; 		\
			(b) = buf[2] << 8;				\
			buf += 3;					\
			break;						\
									\
									\
		default:						\
			fprintf(stderr, 				\
				"Format %d not yet supported\n",	\
				format);				\
	}								\
}						

/* found this in yuv2ppm.c */

int yuv2rgb(int y, int u, int v)
{
   uint32_t pixel32;
   unsigned char *pixel = (unsigned char *)&pixel32;
   int r, g, b;
   
#if 0
   /*
     One formula I found:  (not the right one)
    
     R = 1.164(Y - 16) + 1.596(Cr - 128)
     G = 1.164(Y - 16) - 0.813(Cr - 128) - 0.391(Cb - 128)
     B = 1.164(Y - 16)                   + 2.018(Cb - 128)
   */
      
   r = (1.164 * (y - 16))
      + (2.018 * (v - 128));
   g = (1.164 * (y - 16))
      - (0.813 * (u - 128))
      - (0.391 * (v - 128));
   b = (1.164 * (y - 16))
      + (1.596 * (u - 128));
#else
   /*
     Another formula I found:  (seems to work)
     
     R = Y + 1.370705 (V-128)
     G = Y - 0.698001 (V-128) - 0.337633 (U-128)
     B = Y + 1.732446 (U-128)
   */
   
   r = y + (1.370705 * (v-128));
   g = y - (0.698001 * (v-128)) - (0.337633 * (u-128));
   b = y + (1.732446 * (u-128));
#endif

   /* Even with proper conversion, some values still need clipping. */
   if (r > 255) r = 255;
   if (g > 255) g = 255;
   if (b > 255) b = 255;
   if (r < 0) r = 0;
   if (g < 0) g = 0;
   if (b < 0) b = 0;

   /* Values only go from 0-220..  Why? */
   pixel[0] = r * 220 / 256;
   pixel[1] = g * 220 / 256;
   pixel[2] = b * 220 / 256;
   pixel[3] = 0;

   /* Debug
   printf("yuv2rgb(%i, %i, %i) -> %i, %i, %i  (0x%x)\n",
	  y, u, v,
	  pixel[0], pixel[1], pixel[2],
	  pixel32);
   */
   
   return pixel32;
}

/* end of yuv2ppm.c excerpt */

int get_brightness_adj(unsigned char *image, long size, int *brightness, int d)
{
  long i, tot = 0;
  for (i=0;i<size*d;i++)
    tot += image[i];
  *brightness = (128 - tot/(size*3))/3;
  return !((tot/(size*3)) >= 126 && (tot/(size*3)) <= 130);
}

/* print queried capabilities */
void print_capabilities(struct video_capability *cap, char *devicename)
{
  /* name */
  fprintf(stderr, " capabilities of \"%s\" (%s):\n", cap->name, devicename);
  /* type of device */
  if(cap->type & VID_TYPE_CAPTURE)
    fprintf(stderr, "  VID_TYPE_CAPTURE\n");
  if(cap->type & VID_TYPE_TUNER)
    fprintf(stderr, "  VID_TYPE_TUNER\n");
  if(cap->type & VID_TYPE_TELETEXT)
    fprintf(stderr, "  VID_TYPE_TELETEXT\n");
  if(cap->type & VID_TYPE_OVERLAY)
    fprintf(stderr, "  VID_TYPE_OVERLAY\n");
  if(cap->type & VID_TYPE_CHROMAKEY)
    fprintf(stderr, "  VID_TYPE_CHROMAKEY\n");
  if(cap->type & VID_TYPE_CLIPPING)
    fprintf(stderr, "  VID_TYPE_CLIPPING\n");
  if(cap->type & VID_TYPE_FRAMERAM)
    fprintf(stderr, "  VID_TYPE_FRAMERAM\n");
  if(cap->type & VID_TYPE_SCALES)
    fprintf(stderr, "  VID_TYPE_SCALES\n");
  if(cap->type & VID_TYPE_MONOCHROME)
    fprintf(stderr, "  VID_TYPE_MONOCHROME\n");
  if(cap->type & VID_TYPE_SUBCAPTURE)
    fprintf(stderr, "  VID_TYPE_SUBCAPTURE\n");
  if(cap->type & VID_TYPE_MPEG_DECODER)
    fprintf(stderr, "  VID_TYPE_MPEG_DECODER\n");
  if(cap->type & VID_TYPE_MPEG_ENCODER)
    fprintf(stderr, "  VID_TYPE_MPEG_ENCODER\n");
  if(cap->type & VID_TYPE_MJPEG_DECODER)
    fprintf(stderr, "  VID_TYPE_MJPEG_DECODER\n");
  if(cap->type & VID_TYPE_MJPEG_ENCODER)
    fprintf(stderr, "  VID_TYPE_MJPEG_ENCODER\n");
  /* number of channels */
  fprintf(stderr, " number of channels: %d\n", cap->channels);
  /* number of audio devices */
  fprintf(stderr, " number of audio devices: %d\n", cap->audios);
  /* minimum and maximum image size */
  fprintf(stderr, " image size: ");
  fprintf(stderr, "%d <= width <= %d, %d <= height <= %d\n", cap->minwidth,
                  cap->maxwidth, cap->minheight, cap->maxheight);
}

void print_window_settings(struct video_window *win)
{
  fprintf(stderr, " coordinates: x=%u, y=%u\n", win->x, win->y);
  fprintf(stderr, " size: %ux%u\n", win->width, win->height);
  fprintf(stderr, " chromakey: %u\n", win->chromakey);
}

void print_image_properties(struct video_picture *vpic)
{
  fprintf(stderr, " brightness: %u\n", vpic->brightness);
  fprintf(stderr, " hue:        %u\n", vpic->hue);
  fprintf(stderr, " colour:     %u\n", vpic->colour);
  fprintf(stderr, " contrast:   %u\n", vpic->contrast);
  fprintf(stderr, " whiteness:  %u (only for b/w capturing)\n",vpic->whiteness);
  fprintf(stderr, " depth:      %u\n", vpic->depth);
  fprintf(stderr, " palette:    ");
  switch(vpic->palette) {
    case VIDEO_PALETTE_GREY: fprintf(stderr, "VIDEO_PALETTE_GREY"); break;
    case VIDEO_PALETTE_HI240: fprintf(stderr, "VIDEO_PALETTE_HI240"); break;
    case VIDEO_PALETTE_RGB565: fprintf(stderr, "VIDEO_PALETTE_RGB565"); break;
    case VIDEO_PALETTE_RGB24: fprintf(stderr, "VIDEO_PALETTE_RGB24"); break;
    case VIDEO_PALETTE_RGB32: fprintf(stderr, "VIDEO_PALETTE_RGB32"); break;
    case VIDEO_PALETTE_RGB555: fprintf(stderr, "VIDEO_PALETTE_RGB555"); break;
    case VIDEO_PALETTE_YUV422: fprintf(stderr, "VIDEO_PALETTE_YUV422"); break;
    case VIDEO_PALETTE_YUYV: fprintf(stderr, "VIDEO_PALETTE_YUYV"); break;
    case VIDEO_PALETTE_UYVY: fprintf(stderr, "VIDEO_PALETTE_UYVY"); break;
    case VIDEO_PALETTE_YUV420: fprintf(stderr, "VIDEO_PALETTE_YUV420"); break;
    case VIDEO_PALETTE_YUV411: fprintf(stderr, "VIDEO_PALETTE_YUV411"); break;
    case VIDEO_PALETTE_RAW: fprintf(stderr, "VIDEO_PALETTE_RAW"); break;
    case VIDEO_PALETTE_YUV422P: fprintf(stderr,"VIDEO_PALETTE_YUV422P"); break;
    case VIDEO_PALETTE_YUV411P: fprintf(stderr,"VIDEO_PALETTE_YUV411P"); break;
    case VIDEO_PALETTE_YUV420P: fprintf(stderr,"VIDEO_PALETTE_YUV420P"); break;
    case VIDEO_PALETTE_YUV410P: fprintf(stderr,"VIDEO_PALETTE_YUV410P"); break;
    break;
    default: fprintf(stderr, "unknown palette"); break;
  }
  fprintf(stderr, " (%u)\n", vpic->palette);
}

void print_usage(char *programname)
{
  fprintf(stderr, "Usage: %s [OPTION]...\n", programname);
  fprintf(stderr, "without any options an image is written to stdout using "
                  "current device settings\n");
  fprintf(stderr, "\nOptions:\n");
  fprintf(stderr, "  -h|--help                   print this help\n");
  fprintf(stderr, "  -V|--version                print program version number\n"
          );
  fprintf(stderr, "  -i|--info-only              query capabilities of video "
                  "device and exit\n");
  fprintf(stderr, "  -s|--set-only               set device parameters and "
                  "exit\n");
  fprintf(stderr, "  -f|--file                   write output to file (- for "
                  "stdout)\n");
  fprintf(stderr, "  -d|--device                 specify video4linux device\n");
  fprintf(stderr, "  -v|--verbose                verbose program execution\n");
  fprintf(stderr, "  -q|--quiet                  quiet program execution\n");
  fprintf(stderr, "  -S|--size=WIDTHxHEIGHT      set image WIDTH and HEIGHT\n");
  fprintf(stderr, "  -b|--brightness=BRIGHTNESS  set BRIGHTNESS\n");
  fprintf(stderr, "  -H|--hue=HUE                set HUE\n");
  fprintf(stderr, "  -c|--colour=COLOUR          set COLOUR\n");
  fprintf(stderr, "  -C|--contrast=CONTRAST      set CONTRAST\n");
  fprintf(stderr, "  -w|--whiteness=WHITENESS    set WHITENESS"
                  " (monochrome only)\n");
  fprintf(stderr, "  -p|--palette=PALETTE        set capture image PALETTE\n");
  fprintf(stderr, "                              (--palette=help for list)\n");
  fprintf(stderr, "  -D|--depth=DEPTH            set capture image DEPTH\n");
  fprintf(stderr, "  -a|--auto-bright            use vgrabber's automatic"
                  " RGB brightness adjust\n");
  fprintf(stderr, "  -t|--max-tries=NUMBER       maximum NUMBER of tries for"
                  " RGB brightness adjust\n");
  fprintf(stderr, "  -g|--swap-gb                swap green and blue bytes"
                  " in output\n");
  fprintf(stderr, "\nDefaults:\n");
  fprintf(stderr, "  write output to \"%s\"\n",
          strcmp("-", DEFAULT_OUTFILE) == 0 ? "stdout" : DEFAULT_OUTFILE);
  fprintf(stderr, "  use v4l device \"" DEFAULT_FILE "\" for input\n");
}

void print_version(char *programname)
{
  fprintf(stderr, "WebCam Go image grabber (%s) version " VERSION "\n",
          programname);
  fprintf(stderr, "copyright (C) 2005 by Erik Auerswald\n");
}

void print_palette_help(void)
{
  fprintf(stderr, "PALETTE is one of (case is ignored):\n");
  fprintf(stderr, "GREY, HI240, RGB565, RGB24, RGB32, RGB555, YUV422, YUYV\n");
  fprintf(stderr, "UYVY, YUV420, YUV411, RAW, YUV422P, YUV411P, YUV420P, "
                  "YUV410P\n");
}

int main(int argc, char **argv)
{
  /*** variable definitions ***/
  int fd = -1;
  FILE *ofd = NULL;
  int f;
  char *devicename;
  char *outfilename;
  struct video_capability cap;
  struct video_window win;
  struct video_picture vpic;
  char palettes[17][10] = { {""}, {"GREY"}, {"HI240"}, {"RGB565"}, {"RGB24"},
                      {"RGB32"}, {"RGB555"}, {"YUV422"}, {"YUYV"}, {"UYVY"},
		      {"YUV420"}, {"YUV411"}, {"RAW"}, {"YUV422P"}, {"YUV411P"},
		      {"YUV420P"}, {"YUV410P"} };

  /* RGB and image data variables */
  unsigned char *buffer, *src;
  int /* bpp = 24, */ r, g, b;
  unsigned int i, src_depth;

  /* UYVY to RGB conversion variables */
  uint32_t pixel32 = 0;
  int y=0, u=0, v=0;
  unsigned char *pixels = (unsigned char *)&pixel32;

  /* image format variables */
  int set_image_format = 0;
  int32_t brightness = -1, hue = -1, colour = -1, contrast = -1,
          whiteness = -1, depth = -1, palette = -1;
  uint32_t width=0, height=0;

  /* program operation flags */
  int show_info = 0;
  int set_only = 0;
  int verbose = 0;
  int quiet = 0;
  int auto_bright = 0;
  int swap_gb = 0;
  unsigned int max_tries = MAX_BRIGHTNESS_LOOPS;

  /*** begin of program code ***/
  /* set default devicename and output filename */
  devicename = strdup(DEFAULT_FILE);
  outfilename = strdup(DEFAULT_OUTFILE);

  /* parse command line parameters using getopt */
  while(1) {
    int option_index = 0;
    char c;
    static struct option long_options[] = {
      { "file", 1, 0, 'f' },		/* filename for output */
      { "device", 1, 0, 'd' },		/* v4l device */
      { "info-only", 0, 0, 'i' },	/* show device information */
      { "set-only", 0, 0, 's' },	/* set device parameters */
      { "help", 0, 0, 'h' },		/* print help */
      { "version", 0, 0, 'V' },		/* show version */
      { "verbose", 0, 0, 'v' },		/* verbose execution */
      { "quiet", 0, 0, 'q' },		/* quiet execution */
      { "brightness", 1, 0, 'b' },	/* set brightness */
      { "hue", 1, 0, 'H' },		/* set hue */
      { "colour", 1, 0, 'c' },		/* set colour */
      { "contrast", 1, 0, 'C' },	/* set contrast */
      { "whiteness", 1, 0, 'w' },	/* set whiteness */
      { "depth", 1, 0, 'D' },		/* set capture depth */
      { "palette", 1, 0, 'p' },		/* set palette */
      { "size", 1, 0, 'S' },		/* set image size */
      { "auto-bright", 0, 0, 'a' },	/* use automatic brightness adjust */
      { "max-tries", 1, 0, 't' },	/* max no of tries for brightness adj.*/
      { "swap-gb", 0, 0, 'g' },		/* swap green and blue bytes */
      { 0, 0, 0, 0 }			/* terminate long options */
    };

    c = getopt_long(argc, argv, "f:d:ishVvqb:H:c:C:w:D:p:S:at:g", long_options,
                    &option_index);
    if(c == -1)
      break;

    switch(c) {
      case 'f':
        if(optarg) {
	  free(outfilename);
	  outfilename = strdup(optarg);
	}
	break;

      case 'd':
        if(optarg) {
	  free(devicename);
	  devicename = strdup(optarg);
	}
	break;

      case 'i':
        show_info = 1;
	break;

      case 's':
        set_only = 1;
	break;

      case 'h':
        print_version(argv[0]); fprintf(stderr, "\n");
        print_usage(argv[0]);
	exit(0);
	break;

      case 'V':
        print_version(argv[0]);
	exit(0);
	break;

      case 'v':
        verbose = 1;
	quiet = 0;
	break;

      case 'q':
        quiet = 1;
	verbose = 0;
	break;

      case 'b':
        if(optarg) {
	  brightness = (int32_t) atoi(optarg);
	  set_image_format = 1;
	}
	break;

      case 'H':
        if(optarg) {
	  hue = (int32_t) atoi(optarg);
	  set_image_format = 1;
	}
	break;

      case 'c':
        if(optarg) {
	  colour = (int32_t) atoi(optarg);
	  set_image_format = 1;
	}
	break;

      case 'C':
        if(optarg) {
	  contrast = (int32_t) atoi(optarg);
	  set_image_format = 1;
	}
	break;

      case 'w':
        if(optarg) {
	  whiteness = (int32_t) atoi(optarg);
	  set_image_format = 1;
	}
	break;

      case 'D':
        if(optarg) {
	  depth = (int32_t) atoi(optarg);
	  set_image_format = 1;
	}
	break;

      case 'p':
        if(optarg) {
	  set_image_format = 1;
	  /* accept names for palette */
	  if(strcasecmp("grey", optarg) == 0) {
	    palette = VIDEO_PALETTE_GREY;
	    depth = 8;
	  }
	  else if(strcasecmp("hi240", optarg) == 0) {
	    palette = VIDEO_PALETTE_HI240;
	    depth = 8;
	  }
	  else if(strcasecmp("rgb565", optarg) == 0) {
	    palette = VIDEO_PALETTE_RGB565;
	    depth = 16;
	  }
	  else if(strcasecmp("rgb24", optarg) == 0) {
	    palette = VIDEO_PALETTE_RGB24;
	    depth = 24;
	  }
	  else if(strcasecmp("rgb32", optarg) == 0) {
	    palette = VIDEO_PALETTE_RGB32;
	    depth = 32;
	  }
	  else if(strcasecmp("rgb555", optarg) == 0) {
	    palette = VIDEO_PALETTE_RGB555;
	    depth = 16;
	  }
	  else if(strcasecmp("yuv422", optarg) == 0) {
	    palette = VIDEO_PALETTE_YUV422;
	    depth = 16;
	  }
	  else if(strcasecmp("yuyv", optarg) == 0) {
	    palette = VIDEO_PALETTE_YUYV;
	    depth = 16;
	  }
	  else if(strcasecmp("uyvy", optarg) == 0) {
	    palette = VIDEO_PALETTE_UYVY;
	    depth = 16;
	  }
	  else if(strcasecmp("yuv420", optarg) == 0) {
	    palette = VIDEO_PALETTE_YUV420;
	    depth = 16;	/* this is a guess... */
	  }
	  else if(strcasecmp("yuv411", optarg) == 0) {
	    palette = VIDEO_PALETTE_YUV411;
	    depth = 16;	/* this is a guess... */
	  }
	  else if(strcasecmp("raw", optarg) == 0) {
	    palette = VIDEO_PALETTE_RAW;
	    depth = 8;
	  }
	  else if(strcasecmp("yuv422p", optarg) == 0) {
	    palette = VIDEO_PALETTE_YUV422P;
	    depth = 16;
	  }
	  else if(strcasecmp("yuv411p", optarg) == 0) {
	    palette = VIDEO_PALETTE_YUV411P;
	    depth = 16;
	  }
	  else if(strcasecmp("yuv420p", optarg) == 0) {
	    palette = VIDEO_PALETTE_YUV420P;
	    depth = 16;
	  }
	  else if(strcasecmp("yuv410p", optarg) == 0) {
	    palette = VIDEO_PALETTE_YUV410P;
	    depth = 16;
	  }
	  else {
	    if((strcasecmp("help", optarg) != 0) && (!quiet))
	      fprintf(stderr, "unknown palette %s\n", optarg);
	    print_palette_help();
	    exit(1);
	  }
	}
	break;

      case 'S':
        if(optarg) {
	  char *tmp_s;
	  char *tmp_cp;
	  if((tmp_cp = index(optarg, 'x'))) {
	    tmp_s = strdup(optarg);
	    *tmp_cp = '\0';
	    tmp_cp++;
	    width = atoi(tmp_s);
	    height = atoi(tmp_cp);
	    set_image_format = 1;
	    free(tmp_s);
	  } else {
	    fprintf(stderr, "error: could not parse size argument \"%s\"\n",
	            optarg);
	  }
	}
	break;

      case 'a':
        auto_bright = 1;
        break;

      case 't':
        if(optarg) {
	  int tmp_max_tries = atoi(optarg);
	  if(tmp_max_tries >= 0) {
	    max_tries = tmp_max_tries;
	  } else {
	    fprintf(stderr, "error: number of tries must be positive\n");
	  }
	}
	break;

      case 'g':
        swap_gb = 1;
	break;

      case '?':
        fprintf(stderr, "error: unknown option\n");
        print_usage(argv[0]);
	exit(1);
	break;

      default:	/* should never be reached if getopt() works */
        fprintf(stderr, "error: getopt() returned character 0x%x\n", c);
    }
  }
  if(verbose)
    print_version(argv[0]);

  /* open video device */
  if(verbose)
    fprintf(stderr, "opening video device...\n");
  fd = open(devicename, O_RDONLY);
  if (fd < 0) {
    perror(devicename);
    exit(1);
  }

  /* query device capabilities */
  if(verbose)
    fprintf(stderr, "querying device capabilities...\n");
  if (ioctl(fd, VIDIOCGCAP, &cap) < 0) {
    perror("VIDIOGCAP");
    if(!quiet)
      fprintf(stderr, "(%s not a video4linux device?)\n", devicename);
    close(fd);
    exit(1);
  }

  if(show_info)
    print_capabilities(&cap, devicename);
  free(devicename);	/* not needed any more */

  /* query current window settings */
  if(verbose)
    fprintf(stderr, "querying current video window settings...\n");
  if (ioctl(fd, VIDIOCGWIN, &win) < 0) {
    perror("VIDIOCGWIN");
    close(fd);
    exit(1);
  }
  if(show_info)
    print_window_settings(&win);

  /* query image properties */
  if(verbose)
    fprintf(stderr, "querying current image properties...\n");
  if (ioctl(fd, VIDIOCGPICT, &vpic) < 0) {
    perror("VIDIOCGPICT");
    close(fd);
    exit(1);
  }
  if(show_info) {
    print_image_properties(&vpic);
    exit(0);
  }

  /* set image format */
  if(set_image_format) {
    if(width > 0) {
      if((width >= (uint32_t) cap.minwidth) &&
	 (width <= (uint32_t) cap.maxwidth)) {
	if(verbose)
	  fprintf(stderr, "setting image width %u...\n", width);
	win.width = width;
      } else if(!quiet) {
	fprintf(stderr, "ignoring image width %u not in [%u,%u]\n", width,
	        cap.minwidth, cap.maxwidth);
      }
    }
    if(height > 0) {
      if((height >= (uint32_t) cap.minheight) &&
	 (height <= (uint32_t) cap.maxheight)) {
	if(verbose)
	  fprintf(stderr, "setting image height %u...\n", height);
	win.height = height;
      } else if(!quiet) {
	fprintf(stderr, "ignoring image height %u not in [%u,%u]\n", height,
	        cap.minheight, cap.maxheight);
      }
    }
    if(brightness > -1) {
      if(brightness <= 65535) {
	if(verbose)
	  fprintf(stderr, "setting brightness %u...\n", brightness);
	vpic.brightness = brightness;
      } else if(!quiet) {
	fprintf(stderr, "ignoring brightness %u > 65535\n", brightness);
      }
    }
    if(hue > -1) {
      if(hue <= 65535) {
	if(verbose)
	  fprintf(stderr, "setting hue %u...\n", hue);
	vpic.hue = hue;
      } else if(!quiet) {
	fprintf(stderr, "ignoring hue %u > 65535\n", hue);
      }
    }
    if(colour > -1) {
      if(colour <= 65535) {
	if(verbose)
	  fprintf(stderr, "setting colour %u...\n", colour);
	vpic.colour = colour;
      } else if(!quiet) {
	fprintf(stderr, "ignoring colour %u > 65535\n", colour);
      }
    }
    if(contrast > -1) {
      if(contrast <= 65535) {
	if(verbose)
	  fprintf(stderr, "setting contrast %u...\n", contrast);
	vpic.contrast = contrast;
      } else if(!quiet) {
	fprintf(stderr, "ignoring contrast %u > 65535\n", contrast);
      }
    }
    if(whiteness > -1) {
      if(whiteness <= 65535) {
	if(verbose)
	  fprintf(stderr, "setting whiteness %u...\n", whiteness);
	vpic.whiteness = whiteness;
      } else if(!quiet) {
	fprintf(stderr, "ignoring whiteness %u > 65535\n", whiteness);
      }
    }
    if(depth > -1) {
      if(depth <= 65535) {
	if(verbose)
	  fprintf(stderr, "setting capture depth %u...\n", depth);
	vpic.depth = depth;
      } else if(!quiet) {
	fprintf(stderr, "ignoring capture depth %u > 65535\n", depth);
      }
    }
    if(palette > -1) {
      if(palette <= VIDEO_PALETTE_YUV410P) {	/* highest used palette */
	if(verbose)
	  fprintf(stderr, "setting palette %u...\n", palette);
	vpic.palette = palette;
      } else if(!quiet) {
	fprintf(stderr, "ignoring palette %u > VIDEO_PALETTE_YUV410P\n",
		palette);
      }
    }
    /* set image size */
    if((width > 0) || (height > 0)) {
      if(verbose)
	fprintf(stderr, "setting image size...\n");
      if(ioctl(fd, VIDIOCSWIN, &win) < 0) {
	perror("VIDIOCSWIN");
	if(!quiet)
	  fprintf(stderr, "warning: could not set image size\n");
      } else {
	if(ioctl(fd, VIDIOCGWIN, &win) < 0) {
	  perror("VIDIOCGWIN");
	  if(!quiet)
	    fprintf(stderr, "error: could not read window settings, exiting\n");
	  exit(1);
	} else {
	  if(win.width != width) {
	    if(!quiet)
	      fprintf(stderr, "warning: set image width to %u\n", win.width);
	  } else if(verbose) {
	    fprintf(stderr, "successfully set image width %u\n", win.width);
	  }
	  if(win.height != height) {
	    if(!quiet)
	      fprintf(stderr, "warning: set image height to %u\n", win.height);
	  } else if(verbose) {
	    fprintf(stderr, "successfully set image height %u\n", win.height);
	  }
	}
      }
    }
    /* set image properties */
    if(verbose)
      fprintf(stderr, "setting image properties...\n");
    if(ioctl(fd, VIDIOCSPICT, &vpic) < 0) {
      perror("VIDIOCSPICT");
      if(!quiet)
	fprintf(stderr, "warning: could not set image properties\n");
    } else {	/* check if request could be fulfilled */
      if(ioctl(fd, VIDIOCGPICT, &vpic) < 0) {
	perror("VIDIOCGPICT");
	if(!quiet)
	  fprintf(stderr, "error: could not read image properties, exiting\n");
	close(fd);
	exit(1);
      } else {
	if(brightness > -1) {
	  if(brightness != vpic.brightness) {
	    if(!quiet)
	      fprintf(stderr,"warning: set brightness to %u\n",vpic.brightness);
	  } else if(verbose) {
	    fprintf(stderr, "brightness successfully set to %u\n",
	            vpic.brightness);
	  }
	}
	if(hue > -1) {
	  if(hue != vpic.hue) {
	    if(!quiet)
	      fprintf(stderr,"warning: set hue to %u\n",vpic.hue);
	  } else if(verbose) {
	    fprintf(stderr, "hue successfully set to %u\n",
	            vpic.hue);
	  }
	}
	if(colour > -1) {
	  if(colour != vpic.colour) {
	    if(!quiet)
	      fprintf(stderr,"warning: set colour to %u\n",vpic.colour);
	  } else if(verbose) {
	    fprintf(stderr, "colour successfully set to %u\n",
	            vpic.colour);
	  }
	}
	if(contrast > -1) {
	  if(contrast != vpic.contrast) {
	    if(!quiet)
	      fprintf(stderr,"warning: set contrast to %u\n",vpic.contrast);
	  } else if(verbose) {
	    fprintf(stderr, "contrast successfully set to %u\n",
	            vpic.contrast);
	  }
	}
	if(whiteness > -1) {
	  if(whiteness != vpic.whiteness) {
	    if(!quiet)
	      fprintf(stderr,"warning: set whiteness to %u\n",vpic.whiteness);
	  } else if(verbose) {
	    fprintf(stderr, "whiteness successfully set to %u\n",
	            vpic.whiteness);
	  }
	}
	if(depth > -1) {
	  if(depth != vpic.depth) {
	    if(!quiet)
	      fprintf(stderr,"warning: set depth to %u\n",vpic.depth);
	  } else if(verbose) {
	    fprintf(stderr, "depth successfully set to %u\n",
	            vpic.depth);
	  }
	}
	if(palette > -1) {
	  if(palette != vpic.palette) {
	    if(!quiet)
	      fprintf(stderr,"warning: set palette to %u\n",vpic.palette);
	  } else if(verbose) {
	    fprintf(stderr, "palette successfully set to %u\n",
	            vpic.palette);
	  }
	}
      }
    }
  }
  if(set_only)
    exit(0);

  /* check if a usable palette can be set */
  if(palette < 0) {
    if(verbose)
      fprintf(stderr, "trying to set a supported palette...\n");

    if (cap.type & VID_TYPE_MONOCHROME) {
      vpic.depth=8;
      vpic.palette=VIDEO_PALETTE_GREY;	/* 8bit grey */
      if(ioctl(fd, VIDIOCSPICT, &vpic) < 0) {
	vpic.depth=6;
	if(ioctl(fd, VIDIOCSPICT, &vpic) < 0) {
	  vpic.depth=4;
	  if(ioctl(fd, VIDIOCSPICT, &vpic) < 0) {
	    fprintf(stderr, "Unable to find a supported capture format.\n");
	    close(fd);
	    exit(1);
	  }
	}
      }
    } else {
      vpic.depth=24;
      vpic.palette=VIDEO_PALETTE_RGB24;
      
      if(ioctl(fd, VIDIOCSPICT, &vpic) < 0) {
	vpic.palette=VIDEO_PALETTE_RGB565;
	vpic.depth=16;
	
	if(ioctl(fd, VIDIOCSPICT, &vpic)==-1) {
	  vpic.palette=VIDEO_PALETTE_RGB555;
	  vpic.depth=15;
	  
	  if(ioctl(fd, VIDIOCSPICT, &vpic)==-1) {
	    vpic.palette=VIDEO_PALETTE_UYVY;
	    vpic.depth=16;

	    if(ioctl(fd, VIDIOCSPICT, &vpic)==-1) {
	      vpic.palette=VIDEO_PALETTE_YUYV;
	      vpic.depth=16;

	      if(ioctl(fd, VIDIOCSPICT, &vpic)==-1) {
		fprintf(stderr,
		        "\nUnable to find a supported capture format.\n");
		return -1;
	      }
	    }
	  }
	}
      }
    }
  }
  src_depth = vpic.depth;
  if(verbose) {
    fprintf(stderr, "using palette %s (depth %d bpp)\n", palettes[vpic.palette],
            vpic.depth);
  }
  /* exit if palette can not be converted to image file */
  if(!((vpic.palette == VIDEO_PALETTE_GREY) ||
       (vpic.palette == VIDEO_PALETTE_RGB24) ||
       (vpic.palette == VIDEO_PALETTE_RGB565) ||
       (vpic.palette == VIDEO_PALETTE_RGB555) ||
       (vpic.palette == VIDEO_PALETTE_UYVY) ||
       (vpic.palette == VIDEO_PALETTE_YUYV) )) {
    if(!quiet)
      fprintf(stderr, "Palette %s not supported for capturing.\n",
              palettes[vpic.palette]);
    exit(1);
  }
  
  if(verbose)
    fprintf(stderr, "allocating buffer for image...\n");
  buffer = malloc(win.width * win.height * src_depth/8);
  if (!buffer) {
    fprintf(stderr, "Out of memory.\n");
    exit(1);
  }
  
  /* adjust brightness when told so and using some RGB palette */
  i = 0;
  do {
    int newbright;
    if(verbose)
      fprintf(stderr, "capturing image...\n");
    read(fd, buffer, win.width * win.height * src_depth/8);
    if((auto_bright) && (brightness < 0) &&
      (vpic.palette!=VIDEO_PALETTE_UYVY)&&(vpic.palette!=VIDEO_PALETTE_YUYV)) {
      if(verbose)
	fprintf(stderr, "adjusting brightness (try %d/%d)...\n", i, max_tries);
      f = get_brightness_adj(buffer, win.width * win.height, &newbright,
                             src_depth/8);
      if (f) {
	vpic.brightness += (newbright << 8);
	if(ioctl(fd, VIDIOCSPICT, &vpic)==-1) {
	  perror("VIDIOSPICT");
	  break;
	}
      }
    } else {
      break;
    }
    i++;
  } while (f && (i < max_tries));
  close(fd);

  if(verbose) {
    fprintf(stderr, "writing image in Netpbm PPM \"rawbits\" format"
                    " to file %s...\n",
            strcmp("-", outfilename) == 0 ? "/dev/stdout" : outfilename);
    if(swap_gb)
      fprintf(stderr, "(swapping green and blue bytes...)\n");
  }
  /* open output file */
  if(strcmp("-", outfilename) == 0)
    ofd = stdout;
  else {
    ofd = fopen(outfilename, "w");
    if(ofd == NULL) {
      perror(outfilename);
      exit(1);
    }
  }
    
  /* write ppm header */
  fprintf(ofd, "P6\n%d %d 255\n", win.width, win.height);

  src = buffer;

  for (i = 0; i < win.width * win.height; i++) {
    if((vpic.palette != VIDEO_PALETTE_UYVY) &&
       (vpic.palette != VIDEO_PALETTE_YUYV)) {
      READ_VIDEO_PIXEL(src, vpic.palette, src_depth, r, g, b);
      fputc(r>>8, ofd);
      swap_gb ? fputc(b>>8, ofd) : fputc(g>>8, ofd);
      swap_gb ? fputc(g>>8, ofd) : fputc(b>>8, ofd);
    } else {
      if(vpic.palette == VIDEO_PALETTE_UYVY) {
	if(i==0) {
	  v = src[3];
	}
	if((i%2) == 0) {
	  u = src[0];
	  y = src[1];
	  src += 2;		/* v from next 16bit word */
	} else {
	  v = src[0];
	  y = src[1];
	  src += 2;
	}
      } else if(vpic.palette == VIDEO_PALETTE_YUYV) {
	if(i==0) {
	  v = src[4];
	}
	if((i%2) == 0) {
	  y = src[0];
	  u = src[1];
	  src += 2;		/* v from next 16bit word */
	} else {
	  y = src[0];
	  v = src[1];
	  src += 2;
	}
      }
      pixel32 = yuv2rgb(y, u, v);
      fputc(pixels[0], ofd);
      fputc(pixels[swap_gb ? 2 : 1], ofd);
      fputc(pixels[swap_gb ? 1 : 2], ofd);
    }
  }
  fclose(ofd);
  exit(0);
}
